@@ -8,9 +8,6 @@ use Encode;
 use PDF::API2;
 use Text::Layout;
 use IO::String;
-use Carp;
-
-use App::Music::ChordPro::Output::Common qw( fmt_subst prep_outlines demarkup );
 
 # For regression testing, run perl with PERL_HASH_SEED set to zero.
 # This eliminates the arbitrary order of font definitions and triggers
@@ -23,7 +20,6 @@ my %fontcache;			# speeds up 2 seconds per song
 sub new {
     my ( $pkg, $ps, $pdfapi ) = @_;
     my $self = bless { ps => $ps }, $pkg;
-    $self->{pdfapi} = $pdfapi;
     $self->{pdf} = $pdfapi->new;
     $self->{pdf}->{forcecompress} = 0 if $regtest;
     $self->{pdf}->mediabox( $ps->{papersize}->[0],
@@ -67,34 +63,6 @@ sub wrap {
     return ( $text, $ex );
 }
 
-sub _fgcolor {
-    my ( $self, $col ) = @_;
-    if ( !defined($col) || $col eq "foreground" ) {
-	$col = $self->{ps}->{theme}->{foreground};
-    }
-    elsif ( $col eq "background" ) {
-	$col = $self->{ps}->{theme}->{background};
-    }
-    elsif ( !$col ) {
-	Carp::confess("Undefined fgcolor: $col");
-    }
-    $col;
-}
-
-sub _bgcolor {
-    my ( $self, $col ) = @_;
-    if ( !defined($col) || $col eq "background" ) {
-	$col = $self->{ps}->{theme}->{background};
-    }
-    elsif ( $col eq "foreground" ) {
-	$col = $self->{ps}->{theme}->{foreground};
-    }
-    elsif ( !$col ) {
-	Carp::confess("Undefined bgcolor: $col");
-    }
-    $col;
-}
-
 sub text {
     my ( $self, $text, $x, $y, $font, $size, $nomarkup ) = @_;
 #    print STDERR ("T: @_\n");
@@ -103,12 +71,8 @@ sub text {
 
     $self->{layout}->set_font_description($font->{fd});
     $self->{layout}->set_font_size($size);
-    my $col = $self->_fgcolor($font->{color});
-    if ( $col ne $self->{ps}->{theme}->{foreground} ) {
-	$text = "<span color='" . $col . "'>" . $text . "</span>";
-    }
-    else {
-	$self->{layout}->{_currentcolor} = $col;
+    if ( $font->{color} && $font->{color} ne "black" ) {
+	$text = "<span color='" . $font->{color} . "'>" . $text . "</span>";
     }
     if ( $nomarkup ) {
 	$self->{layout}->set_text($text);
@@ -131,7 +95,7 @@ sub text {
 	#printf("BB: %.2f %.2f %.2f %.2f\n", @{$e}{qw( x y width height ) } );
 	# Draw background and.or frame.
 	my $d = $debug ? 0 : 1;
-	$frame = $debug || $font->{color} || $self->{ps}->{theme}->{foreground} if $frame;
+	$frame = $debug || $font->{color} || "black" if $frame;
 	$self->rectxy( $x + $e->{x} - $d,
 		       $y + $e->{y} + $d,
 		       $x + $e->{x} + $e->{width} + $d,
@@ -153,12 +117,8 @@ sub text_nobl {
 
     $self->{layout}->set_font_description($font->{fd});
     $self->{layout}->set_font_size($size);
-    my $col = $self->_fgcolor($font->{color});
-    if ( $col ne $self->{ps}->{theme}->{foreground} ) {
-	$text = "<span color='" . $col . "'>" . $text . "</span>";
-    }
-    else {
-	$self->{layout}->{_currentcolor} = $col;
+    if ( $font->{color} && $font->{color} ne "black" ) {
+	$text = "<span color='" . $font->{color} . "'>" . $text . "</span>";
     }
     $self->{layout}->set_markup($text);
     $self->{layout}->show( $x, $y, $self->{pdftext} );
@@ -175,7 +135,7 @@ sub text_nobl {
 	#printf("BB: %.2f %.2f %.2f %.2f\n", @{$e}{qw( x y width height ) } );
 	# Draw background and.or frame.
 	my $d = $debug ? 0 : 1;
-	$frame = $debug || $font->{color} || $self->{ps}->{theme}->{foreground} if $frame;
+	$frame = $debug || $font->{color} || "black" if $frame;
 	$self->rectxy( $x + $e->{x} - $d,
 		       $y + $e->{y} + $d,
 		       $x + $e->{x} + $e->{width} + $d,
@@ -191,9 +151,7 @@ sub text_nobl {
 sub setfont {
     my ( $self, $font, $size ) = @_;
     $self->{font} = $font;
-    warn("PDF: Font ", $font->{_ff}, " should have a size!\n")
-      unless $size ||= $font->{size};
-    $self->{fontsize} = $size ||= $font->{size} || $font->{fd}->{size};
+    $self->{fontsize} = $size ||= $font->{size};
     $self->{pdftext}->font( $font->{fd}->{font}, $size );
 }
 
@@ -225,7 +183,7 @@ sub line {
     my ( $self, $x0, $y0, $x1, $y1, $lw, $color ) = @_;
     my $gfx = $self->{pdfgfx};
     $gfx->save;
-    $gfx->strokecolor( $self->_fgcolor($color) );
+    $gfx->strokecolor($color ||= "black");
     $gfx->linecap(1);
     $gfx->linewidth($lw||1);
     $gfx->move( $x0, $y0 );
@@ -238,7 +196,7 @@ sub hline {
     my ( $self, $x, $y, $w, $lw, $color ) = @_;
     my $gfx = $self->{pdfgfx};
     $gfx->save;
-    $gfx->strokecolor( $self->_fgcolor($color) );
+    $gfx->strokecolor($color ||= "black");
     $gfx->linecap(2);
     $gfx->linewidth($lw||1);
     $gfx->move( $x, $y );
@@ -251,7 +209,7 @@ sub vline {
     my ( $self, $x, $y, $h, $lw, $color ) = @_;
     my $gfx = $self->{pdfgfx};
     $gfx->save;
-    $gfx->strokecolor( $self->_fgcolor($color) );
+    $gfx->strokecolor($color ||= "black");
     $gfx->linecap(2);
     $gfx->linewidth($lw||1);
     $gfx->move( $x, $y );
@@ -264,8 +222,8 @@ sub rectxy {
     my ( $self, $x, $y, $x1, $y1, $lw, $fillcolor, $strokecolor ) = @_;
     my $gfx = $self->{pdfgfx};
     $gfx->save;
-    $gfx->strokecolor($self->_fgcolor($strokecolor)) if $strokecolor;
-    $gfx->fillcolor($self->_fgcolor($fillcolor)) if $fillcolor;
+    $gfx->strokecolor($strokecolor) if $strokecolor;
+    $gfx->fillcolor($fillcolor) if $fillcolor;
     $gfx->linecap(2);
     $gfx->linewidth($lw||1);
     $gfx->rectxy( $x, $y, $x1, $y1 );
@@ -275,29 +233,12 @@ sub rectxy {
     $gfx->restore;
 }
 
-sub poly {
-    my ( $self, $points, $lw, $fillcolor, $strokecolor ) = @_;
-    undef $strokecolor unless $lw;
-    my $gfx = $self->{pdfgfx};
-    $gfx->save;
-    $gfx->strokecolor($self->_fgcolor($strokecolor)) if $strokecolor;
-    $gfx->fillcolor($self->_fgcolor($fillcolor)) if $fillcolor;
-    $gfx->linecap(2);
-    $gfx->linewidth($lw);
-    $gfx->poly( @$points );
-    $gfx->close;
-    $gfx->fill if $fillcolor && !$strokecolor;
-    $gfx->fillstroke if $fillcolor && $strokecolor;
-    $gfx->stroke if $strokecolor && !$fillcolor;
-    $gfx->restore;
-}
-
 sub circle {
     my ( $self, $x, $y, $r, $lw, $fillcolor, $strokecolor ) = @_;
     my $gfx = $self->{pdfgfx};
     $gfx->save;
-    $gfx->strokecolor($self->_fgcolor($strokecolor)) if $strokecolor;
-    $gfx->fillcolor($self->_fgcolor($fillcolor)) if $fillcolor;
+    $gfx->strokecolor($strokecolor) if $strokecolor;
+    $gfx->fillcolor($fillcolor) if $fillcolor;
     $gfx->linewidth($lw||1);
     $gfx->circle( $x, $y, $r );
     $gfx->fill if $fillcolor;
@@ -309,7 +250,7 @@ sub cross {
     my ( $self, $x, $y, $r, $lw, $strokecolor ) = @_;
     my $gfx = $self->{pdfgfx};
     $gfx->save;
-    $gfx->strokecolor($self->_fgcolor($strokecolor)) if $strokecolor;
+    $gfx->strokecolor($strokecolor) if $strokecolor;
     $gfx->linewidth($lw||1);
     $r = 0.9 * $r;
     $gfx->move( $x-$r, $y-$r );
@@ -371,17 +312,6 @@ sub newpage {
 				$ps->{papersize}->[1] );
     $self->{pdfgfx}  = $self->{pdfpage}->gfx;
     $self->{pdftext} = $self->{pdfpage}->text;
-    unless ($ps->{theme}->{background} =~ /^white|none|#ffffff$/i ) {
-	for ( $self->{pdfgfx} ) {
-	    $_->save;
-	    $_->fillcolor( $ps->{theme}->{background} );
-	    $_->linewidth(0);
-	    $_->rectxy( 0, 0, $ps->{papersize}->[0],
-			$ps->{papersize}->[1] );
-	    $_->fill;
-	    $_->restore;
-	}
-    }
 }
 
 sub pagelabel {
@@ -392,6 +322,16 @@ sub pagelabel {
     $self->{pdf}->pageLabel( $page, $opts );
 }
 
+# Substitute %X sequences in title formats.
+sub fmt_subst {
+    goto \&App::Music::ChordPro::Output::Common::fmt_subst;
+}
+
+# Prepare outlines.
+sub prep_outlines {
+    goto \&App::Music::ChordPro::Output::Common::prep_outlines;
+}
+
 sub make_outlines {
     my ( $self, $book, $start ) = @_;
     return unless $book && @$book; # unlikely
@@ -408,18 +348,9 @@ sub make_outlines {
 	# Seems not to matter whether we re-use the root or create new.
 	$ol_root //= $pdf->outlines;
 
-	my $outline;
-
-	# Skip level for a single outline.
-	if ( @{ $self->{ps}->{outlines} } == 1 ) {
-	    $outline = $ol_root;
-	    $outline->closed if $ctl->{collapse}; # TODO?
-	}
-	else {
-	    $outline = $ol_root->outline;
-	    $outline->title( $ctl->{label} );
-	    $outline->closed if $ctl->{collapse};
-	}
+	my $outline = $ol_root->outline;
+	$outline->title( $ctl->{label} );
+	$outline->closed if $ctl->{collapse};
 
 	my %lh;			# letter hierarchy
 	for ( @$book ) {
@@ -454,7 +385,7 @@ sub make_outlines {
 		    $ol = $outline->outline;
 		}
 		# Display info.
-		$ol->title( demarkup( fmt_subst( $song, $ctl->{line} ) ) );
+		$ol->title( fmt_subst( $song, $ctl->{line} ) );
 		$ol->dest($pdf->openpage( $song->{meta}->{tocpage} + $start ));
 	    }
 	}
@@ -481,20 +412,8 @@ sub init_fonts {
 
     my $fc = Text::Layout::FontConfig->new;
 
-    # Add font dirs.
-    my @d = ( @{$ps->{fontdir}}, ::rsc_or_file("fonts/"), $ENV{FONTDIR} );
-    # Avoid rsc result if dummy.
-    splice( @d, -2, 1 ) if $d[-2] eq "fonts/";
-    for my $fontdir ( @d ) {
-	next unless $fontdir;
-	if ( -d $fontdir ) {
-	    $self->{pdfapi}->can("addFontDirs")->($fontdir);
-	    $fc->add_fontdirs($fontdir);
-	}
-	else {
-	    warn("PDF: Ignoring fontdir $fontdir [$!]\n");
-	    undef $fontdir;
-	}
+    if ( $ps->{fontdir} ) {
+	$fc->add_fontdirs( @{ $ps->{fontdir} } );
     }
 
     foreach my $ff ( keys( %{ $ps->{fontconfig} } ) ) {
@@ -523,13 +442,13 @@ sub init_font {
     my ( $self, $ff ) = @_;
     my $ps = $self->{ps};
     my $fd;
-    if ( $ps->{fonts}->{$ff}->{file} ) {
+    if ( !$fd && $ps->{fonts}->{$ff}->{file} ) {
 	$fd = $self->init_filefont($ff);
     }
-    elsif ( $ps->{fonts}->{$ff}->{description} ) {
+    if ( !$fd && $ps->{fonts}->{$ff}->{description} ) {
 	$fd = $self->init_pangofont($ff);
     }
-    elsif ( $ps->{fonts}->{$ff}->{name} ) {
+    if ( !$fd && $ps->{fonts}->{$ff}->{name} ) {
 	$fd = $self->init_corefont($ff);
     }
     warn("No font found for \"$ff\"\n") unless $fd;
@@ -549,7 +468,6 @@ sub init_pangofont {
 	$font->{fd}->{font}->{Name}->{val} =~ s/~.*/~$faketime/ if $regtest;
 	$font->{_ff} = $ff;
 	$font->{fd}->set_shaping( $font->{fd}->get_shaping || $font->{shaping}//0);
-	$font->{size} = $font->{fd}->get_size if $font->{fd}->get_size;
     };
     $font->{fd};
 }
@@ -562,11 +480,10 @@ sub init_filefont {
 
     my $fc = Text::Layout::FontConfig->new;
     eval {
-	my $t = $fc->from_filename($font->{file});
-	$t->get_font($self->{layout}); # force load
-	$t->{font}->{Name}->{val} =~ s/~.*/~$faketime/ if $regtest;
-	$t->{_ff} = $ff;
-	$font->{fd} = $t;
+	$font->{fd} = $fc->from_filename($font->{file});
+	$font->{fd}->get_font($self->{layout}); # force load
+	$font->{fd}->{font}->{Name}->{val} =~ s/~.*/~$faketime/ if $regtest;
+	$font->{_ff} = $ff;
     };
     $font->{fd};
 }
@@ -596,14 +513,13 @@ use File::Temp;
 my $cname;
 sub embed {
     my ( $self, $file ) = @_;
-    return unless -f $file;
     my $a = $self->{pdfpage}->annotation();
 
     # The only reliable way currently is pretend it's a movie :) .
     $a->movie($file, "ChordPro" );
     $a->open(1);
 
-    # Create/reuse temp file for (final) config and run time info.
+    # Create/reuse temp file for (final) config.
     my $cf;
     if ( $cname ) {
 	open( $cf, '>', $cname );
@@ -612,21 +528,13 @@ sub embed {
 	( $cf, $cname ) = File::Temp::tempfile( UNLINK => 0);
     }
     binmode( $cf, ':utf8' );
-    print $cf App::Music::ChordPro::Config::config_final();
+    my $pp = JSON::PP->new->utf8(0)->canonical->indent(4)->pretty;
+    print $cf $pp->encode( { %$::config } );
     close($cf);
 
     $a = $self->{pdfpage}->annotation();
     $a->movie($cname, "ChordProConfig" );
     $a->open(0);
-
-    open( $cf, '>', $cname );
-    binmode( $cf, ':utf8' );
-    print $cf (::runtimeinfo());
-    close($cf);
-
-    $a = $self->{pdfpage}->annotation();
-    $a->movie($cname, "ChordProRunTime" );
-    $a->open(0);
 }
 
 END {
